@@ -1,7 +1,7 @@
 /* 
- * Copyright (C) 2000-2001 the xine project
+ * Copyright (C) 2000-2003 the xine project
  * 
- * This file is part of xine, a unix video player.
+ * This file is part of xine, a free video player.
  * 
  * xine is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
  *
- * $Id: input_vcd.c,v 1.46 2002/07/05 17:32:02 mroi Exp $
+ * $Id: input_vcd.c,v 1.77 2004/12/12 22:01:07 mroi Exp $
  *
  */
 
@@ -50,28 +50,7 @@
 #include "xine_internal.h"
 #include "xineutils.h"
 #include "input_plugin.h"
-
-extern int errno;
-
-#ifdef __GNUC__
-#define LOG_MSG_STDERR(xine, message, args...) {                     \
-    xine_log(xine, XINE_LOG_MSG, message, ##args);                 \
-    fprintf(stderr, message, ##args);                                \
-  }
-#define LOG_MSG(xine, message, args...) {                            \
-    xine_log(xine, XINE_LOG_MSG, message, ##args);                 \
-    printf(message, ##args);                                         \
-  }
-#else
-#define LOG_MSG_STDERR(xine, ...) {                                  \
-    xine_log(xine, XINE_LOG_MSG, __VA_ARGS__);                     \
-    fprintf(stderr, __VA_ARGS__);                                    \
-  }
-#define LOG_MSG(xine, ...) {                                         \
-    xine_log(xine, XINE_LOG_MSG, __VA_ARGS__);                     \
-    printf(__VA_ARGS__);                                             \
-  }
-#endif
+#include "media_helper.h"
 
 #if defined(__sun)
 #define	CDROM	       "/vol/dev/aliases/cdrom0"
@@ -99,14 +78,18 @@ typedef struct {
 
 typedef struct {
 
-  input_plugin_t         input_plugin;
-  
+  input_class_t          input_class;
+
   xine_t                *xine;
 
-  char                  *mrl;
-  config_values_t       *config;
+  /* FIXME: add thread/mutex stuff to protect against multiple instantiation */
 
-  int                    fd;
+  const char            *device;
+
+  char                  *filelist[100];
+
+  int                    mrls_allocated_entries;
+  xine_mrl_t           **mrls;
 
 #if defined (__linux__) || defined(__sun)
   struct cdrom_tochdr    tochdr;
@@ -114,47 +97,59 @@ typedef struct {
 #elif defined (__FreeBSD__)
   struct ioc_toc_header  tochdr;
   struct cd_toc_entry    *tocent;
-  off_t                  cur_sector;
+  off_t                  cur_sec;
 #endif
+
   int                    total_tracks;
+
+} vcd_input_class_t;
+
+typedef struct {
+
+  input_plugin_t         input_plugin;
+  
+  vcd_input_class_t     *cls;
+  
+  xine_stream_t         *stream;
+  
+  char                  *mrl;
+  config_values_t       *config;
+
+  int                    fd;
+
   int                    cur_track;
 
-#if defined (__linux__) || defined(__sun)
+#if defined (__linux__) || defined(__sun) || defined(__FreeBSD__)
   uint8_t                cur_min, cur_sec, cur_frame;
 #endif
 
-  const char            *device;
-
-  char                  *filelist[100];
-
-  int                    mrls_allocated_entries;
-  mrl_t                **mrls;
-
 #if defined(__sun)
   int			 controller_type;
 #endif
 } vcd_input_plugin_t;
 
 
+
 /* ***************************************************************** */
 /*                        Private functions                          */
 /* ***************************************************************** */
 /*
  * Callback for configuratoin changes.
  */
-static void device_change_cb(void *data, cfg_entry_t *cfg) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) data;
+static void device_change_cb (void *data, xine_cfg_entry_t *cfg) {
+  vcd_input_class_t *this = (vcd_input_class_t *) data;
   
   this->device = cfg->str_value;
 }
 
 #if defined (__linux__) || defined(__sun)
-static int input_vcd_read_toc (vcd_input_plugin_t *this) {
+static int input_vcd_read_toc (vcd_input_class_t *this, int fd) {
   int i;
 
   /* read TOC header */
-  if ( ioctl(this->fd, CDROMREADTOCHDR, &this->tochdr) == -1 ) {
-    LOG_MSG_STDERR(this->xine, _("input_vcd : error in ioctl CDROMREADTOCHDR\n"));
+  if ( ioctl(fd, CDROMREADTOCHDR, &this->tochdr) == -1 ) {
+    xprintf (this->xine, XINE_VERBOSITY_DEBUG, 
+	     "input_vcd : error in ioctl CDROMREADTOCHDR\n");
     return -1;
   }
 
@@ -162,9 +157,9 @@ static int input_vcd_read_toc (vcd_input_plugin_t *this) {
   for (i=this->tochdr.cdth_trk0; i<=this->tochdr.cdth_trk1; i++) {
     this->tocent[i-1].cdte_track = i;
     this->tocent[i-1].cdte_format = CDROM_MSF;
-    if ( ioctl(this->fd, CDROMREADTOCENTRY, &this->tocent[i-1]) == -1 ) {
-      LOG_MSG_STDERR(this->xine, 
-		     _("input_vcd: error in ioctl CDROMREADTOCENTRY for track %d\n"), i);
+    if ( ioctl(fd, CDROMREADTOCENTRY, &this->tocent[i-1]) == -1 ) {
+      xprintf (this->xine, XINE_VERBOSITY_DEBUG, 
+	       "input_vcd: error in ioctl CDROMREADTOCENTRY for track %d\n", i);
       return -1;
     }
   }
@@ -173,9 +168,10 @@ static int input_vcd_read_toc (vcd_input_plugin_t *this) {
   this->tocent[this->tochdr.cdth_trk1].cdte_track = CDROM_LEADOUT;
   this->tocent[this->tochdr.cdth_trk1].cdte_format = CDROM_MSF;
 
-  if (ioctl(this->fd, CDROMREADTOCENTRY, 
+  if (ioctl(fd, CDROMREADTOCENTRY, 
 	    &this->tocent[this->tochdr.cdth_trk1]) == -1 ) {
-    LOG_MSG_STDERR(this->xine, _("input_vcd: error in ioctl CDROMREADTOCENTRY for lead-out\n"));
+    xprintf (this->xine, XINE_VERBOSITY_DEBUG, 
+	     "input_vcd: error in ioctl CDROMREADTOCENTRY for lead-out\n");
     return -1;
   }
 
@@ -184,14 +180,15 @@ static int input_vcd_read_toc (vcd_input_plugin_t *this) {
   return 0;
 }
 #elif defined (__FreeBSD__)
-static int input_vcd_read_toc (vcd_input_plugin_t *this) {
+static int input_vcd_read_toc (vcd_input_class_t *this, int fd) {
 
   struct ioc_read_toc_entry te;
   int ntracks;
 
   /* read TOC header */
-  if ( ioctl(this->fd, CDIOREADTOCHEADER, &this->tochdr) == -1 ) {
-    LOG_MSG_STDERR(this->xine, _("input_vcd : error in ioctl CDROMREADTOCHDR\n"));
+  if ( ioctl(fd, CDIOREADTOCHEADER, &this->tochdr) == -1 ) {
+    xprintf (this->xine, XINE_VERBOSITY_DEBUG, 
+	     "input_vcd : error in ioctl CDROMREADTOCHDR\n");
     return -1;
   }
 
@@ -205,8 +202,9 @@ static int input_vcd_read_toc (vcd_input_plugin_t *this) {
   te.data_len = ntracks * sizeof(struct cd_toc_entry);
   te.data = this->tocent;
   
-  if ( ioctl(this->fd, CDIOREADTOCENTRYS, &te) == -1 ){
-    LOG_MSG_STDERR(this->xine, _("input_vcd: error in ioctl CDROMREADTOCENTRY\n"));
+  if ( ioctl(fd, CDIOREADTOCENTRYS, &te) == -1 ){
+    xprintf (this->xine, XINE_VERBOSITY_DEBUG, 
+	     "input_vcd: error in ioctl CDROMREADTOCENTRY\n");
     return -1;
   }
 
@@ -235,8 +233,9 @@ static int sun_vcd_read(vcd_input_plugin_t *this, long lba, cdsector_t *data)
    * ATAPI devices.
    */
   if (this->controller_type == 0) {
-    if (ioctl(this->fd, DKIOCINFO, &cinfo) == 0
-	&& strcmp(cinfo.dki_cname, "ide") == 0)
+    if ( ioctl(this->fd, DKIOCINFO, &cinfo) == 0
+	 && ((strcmp(cinfo.dki_cname, "ide") == 0) 
+	     || (strncmp(cinfo.dki_cname, "pci", 3) == 0)) )
       this->controller_type = SUN_CTRL_ATAPI;
     else
       this->controller_type = SUN_CTRL_SCSI;    
@@ -252,8 +251,8 @@ static int sun_vcd_read(vcd_input_plugin_t *this, long lba, cdsector_t *data)
       cdxa.cdxa_format = CDROM_XA_SECTOR_DATA;
   
       if(ioctl(this->fd,CDROMCDXA,&cdxa)==-1) {
-	  perror("CDROMCDXA");
-	  return -1;
+	xprintf(this->cls->xine, XINE_VERBOSITY_DEBUG, "CDROMCDXA: %s\n", strerror(errno));
+	return -1;
       }
     }
 #else
@@ -264,8 +263,8 @@ static int sun_vcd_read(vcd_input_plugin_t *this, long lba, cdsector_t *data)
       cdread.cdread_buflen = 2336;
 
       if(ioctl(this->fd,CDROMREADMODE2,&cdread)==-1) {
-	  perror("CDROMREADMODE2");
-	  return -1;
+	xprintf(this->cls->xine, XINE_VERBOSITY_DEBUG, "CDROMREADMODE2: %s\n", strerror(errno));
+	return -1;
       }
     }
 #endif
@@ -320,11 +319,12 @@ static int sun_vcd_read(vcd_input_plugin_t *this, long lba, cdsector_t *data)
       sc.uscsi_flags = USCSI_ISOLATE | USCSI_READ;
       sc.uscsi_timeout = 20;
       if (ioctl(this->fd, USCSICMD, &sc)) {
-	perror("USCSICMD: READ CD");
+	xprintf(this->cls->xine, XINE_VERBOSITY_DEBUG, "USCSICMD: READ CD: %s\n", strerror(errno));
 	return -1;
       }
       if (sc.uscsi_status) {
-	LOG_MSG_STDERR(this->xine, _("scsi command failed with status %d\n"), sc.uscsi_status);
+	xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, 
+		 "scsi command failed with status %d\n", sc.uscsi_status);
 	return -1;
       }
     }
@@ -337,73 +337,6 @@ static int sun_vcd_read(vcd_input_plugin_t *this, long lba, cdsector_t *data)
 /*                         END OF PRIVATES                           */
 /* ***************************************************************** */
 
-
-/*
- *
- */
-static int vcd_plugin_open (input_plugin_t *this_gen, char *mrl) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
-  char *filename;
-
-  this->mrl = mrl;
-
-  if (strncasecmp (mrl, "vcd://",6))
-    return 0;
-    
-  this->fd = open (this->device, O_RDONLY);
-
-  if (this->fd == -1) {
-    return 0;
-  }
-
-  if (input_vcd_read_toc (this)) {
-    close (this->fd);
-    this->fd = -1;
-    return 0;
-  }
-
-  filename = (char *) &mrl[6];
-
-  if (sscanf (filename, "%d", &this->cur_track) != 1) {
-    LOG_MSG_STDERR(this->xine, _("input_vcd: malformed MRL. Use vcd://<track #>\n"));
-    close (this->fd);
-    this->fd = -1;
-    return 0;
-  }
-
-  if (this->cur_track>=this->total_tracks) {
-    LOG_MSG_STDERR(this->xine, _("input_vcd: invalid track %d (valid range: 0 .. %d)\n"),
-	    this->cur_track, this->total_tracks-1);
-    close (this->fd);
-    this->fd = -1;
-    return 0;
-  }
-
-#if defined (__linux__) || defined(__sun)
-  this->cur_min   = this->tocent[this->cur_track].cdte_addr.msf.minute;
-  this->cur_sec   = this->tocent[this->cur_track].cdte_addr.msf.second;
-  this->cur_frame = this->tocent[this->cur_track].cdte_addr.msf.frame;
-#elif defined (__FreeBSD__)
-  {
-    int bsize = 2352;
-    if (ioctl (this->fd, CDRIOCSETBLOCKSIZE, &bsize) == -1) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: error in CDRIOCSETBLOCKSIZE %d\n"), errno);
-      return 0;
-    }
-  
-    this->cur_sector = 
-      ntohl(this->tocent
-	    [this->cur_track+1 - this->tochdr.starting_track].addr.lba);
-    
-  }
-#endif
-  
-  return 1;
-}
-
-/*
- *
- */
 #if defined (__linux__)
 static off_t vcd_plugin_read (input_plugin_t *this_gen, 
 				char *buf, off_t nlen) {
@@ -418,7 +351,8 @@ static off_t vcd_plugin_read (input_plugin_t *this_gen,
 
   do
   {  
-    end_msf = &this->tocent[this->cur_track+1].cdte_addr.msf;
+    end_msf = (struct cdrom_msf0 *)
+      &this->cls->tocent[this->cur_track+1].cdte_addr.msf;
 
     /*
     printf ("cur: %02d:%02d:%02d  end: %02d:%02d:%02d\n",
@@ -437,7 +371,8 @@ static off_t vcd_plugin_read (input_plugin_t *this_gen,
     memcpy (&data, &msf, sizeof (msf));
 
     if (ioctl (this->fd, CDROMREADRAW, &data) == -1) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: error in CDROMREADRAW\n"));
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, 
+	       "input_vcd: error in CDROMREADRAW\n");
       return 0;
     }
 
@@ -471,15 +406,15 @@ static off_t vcd_plugin_read (input_plugin_t *this_gen,
     return 0;
 
   do {
-    if (lseek (this->fd, this->cur_sector * bsize, SEEK_SET) == -1) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: seek error %d\n"), errno);
+    if (lseek (this->fd, this->cur_sec * bsize, SEEK_SET) == -1) {
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: seek error %d\n", errno);
       return 0;
     }
     if (read (this->fd, &data, bsize) == -1) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: read error %d\n"), errno);
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: read error %d\n", errno);
       return 0;
     }
-    this->cur_sector++;
+    this->cur_sec++;
   } while ((data.subheader[2]&~0x01)==0x60);
   memcpy (buf, data.data, VCDSECTORSIZE);
   return VCDSECTORSIZE;
@@ -498,7 +433,8 @@ static off_t vcd_plugin_read (input_plugin_t *this_gen,
 
   do
   {  
-    end_msf = &this->tocent[this->cur_track+1].cdte_addr.msf;
+    end_msf = (struct cdrom_msf0 *) 
+      &this->cls->tocent[this->cur_track+1].cdte_addr.msf;
 
     /*
     printf ("cur: %02d:%02d:%02d  end: %02d:%02d:%02d\n",
@@ -513,7 +449,7 @@ static off_t vcd_plugin_read (input_plugin_t *this_gen,
     lba = (this->cur_min * 60 + this->cur_sec) * 75L + this->cur_frame;
 
     if (sun_vcd_read(this, lba, &data) < 0) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: read data failed\n"));
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: read data failed\n");
       return 0;
     }
 
@@ -537,9 +473,6 @@ static off_t vcd_plugin_read (input_plugin_t *this_gen,
 }
 #endif
 
-/*
- *
- */
 #if defined (__linux__)
 static buf_element_t *vcd_plugin_read_block (input_plugin_t *this_gen, 
 					     fifo_buffer_t *fifo, off_t nlen) {
@@ -555,7 +488,7 @@ static buf_element_t *vcd_plugin_read_block (input_plugin_t *this_gen,
 
   do
   {  
-    end_msf = &this->tocent[this->cur_track+1].cdte_addr.msf;
+    end_msf = &this->cls->tocent[this->cur_track+1].cdte_addr.msf;
 
     /*
     printf ("cur: %02d:%02d:%02d  end: %02d:%02d:%02d\n",
@@ -574,7 +507,7 @@ static buf_element_t *vcd_plugin_read_block (input_plugin_t *this_gen,
     memcpy (&data, &msf, sizeof (msf));
 
     if (ioctl (this->fd, CDROMREADRAW, &data) == -1) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: error in CDROMREADRAW\n"));
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: error in CDROMREADRAW\n");
       return NULL;
     }
 
@@ -613,15 +546,15 @@ static buf_element_t *vcd_plugin_read_block (input_plugin_t *this_gen,
     return NULL;
 
   do {
-    if (lseek (this->fd, this->cur_sector * bsize, SEEK_SET) == -1) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: seek error %d\n"), errno);
+    if (lseek (this->fd, this->cur_sec * bsize, SEEK_SET) == -1) {
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: seek error %d\n", errno);
       return NULL;
     }
     if (read (this->fd, &data, bsize) == -1) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: read error %d\n"), errno);
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: read error %d\n", errno);
       return NULL;
     }
-    this->cur_sector++;
+    this->cur_sec++;
   } while ((data.subheader[2]&~0x01)==0x60);
 
   buf = fifo->buffer_pool_alloc (fifo);
@@ -645,7 +578,8 @@ static buf_element_t *vcd_plugin_read_block (input_plugin_t *this_gen,
 
   do
   {  
-    end_msf = &this->tocent[this->cur_track+1].cdte_addr.msf;
+    end_msf = (struct cdrom_msf0 *) 
+      &this->cls->tocent[this->cur_track+1].cdte_addr.msf;
 
     /*
     printf ("cur: %02d:%02d:%02d  end: %02d:%02d:%02d\n",
@@ -660,7 +594,7 @@ static buf_element_t *vcd_plugin_read_block (input_plugin_t *this_gen,
     lba = (this->cur_min * 60 + this->cur_sec) * 75L + this->cur_frame;
 
     if (sun_vcd_read (this, lba, &data) < 0) {
-      LOG_MSG_STDERR(this->xine, _("input_vcd: read data failed\n"));
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: read data failed\n");
       return NULL;
     }
 
@@ -702,7 +636,7 @@ static void vcd_offset_to_time (off_t offset, uint8_t *min, uint8_t *sec,
 
 }
 
-#if defined (__linux__) || defined(__sun)
+#if defined (__linux__) || defined(__sun) || defined(HOST_OS_DARWIN)
 static off_t vcd_plugin_seek (input_plugin_t *this_gen, 
 			      off_t offset, int origin) {
 
@@ -710,7 +644,8 @@ static off_t vcd_plugin_seek (input_plugin_t *this_gen,
   struct cdrom_msf0   *start_msf;
   off_t                sector_pos;
 
-  start_msf = &this->tocent[this->cur_track].cdte_addr.msf;
+  start_msf = (struct cdrom_msf0 *)
+    &this->cls->tocent[this->cur_track].cdte_addr.msf;
 
   switch (origin) {
   case SEEK_SET:
@@ -731,7 +666,8 @@ static off_t vcd_plugin_seek (input_plugin_t *this_gen,
     break;
   case SEEK_CUR:
     if (offset) 
-      LOG_MSG_STDERR(this->xine, _("input_vcd: SEEK_CUR not implemented for offset != 0\n"));
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, 
+	       "input_vcd: SEEK_CUR not implemented for offset != 0\n");
 
     /*
     printf ("input_vcd: current pos: %02d:%02d:%02d\n",
@@ -749,8 +685,8 @@ static off_t vcd_plugin_seek (input_plugin_t *this_gen,
 
     break;
   default:
-    LOG_MSG_STDERR(this->xine, _("input_vcd: error seek to origin %d not implemented!\n"),
-	     origin);
+    xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, 
+	     "input_vcd: error seek to origin %d not implemented!\n", origin);
     return 0;
   }
 
@@ -767,8 +703,8 @@ static off_t vcd_plugin_seek (input_plugin_t *this_gen,
   off_t sector_pos;
 
   start = 
-    ntohl(this->tocent
-	  [this->cur_track+1 - this->tochdr.starting_track].addr.lba);
+    ntohl(this->cls->tocent
+	  [this->cur_track+1 - this->cls->tochdr.starting_track].addr.lba);
 
   /*  printf("seek: start sector:%lu, origin: %d, offset:%qu\n", 
 	 start, origin, offset);
@@ -777,21 +713,21 @@ static off_t vcd_plugin_seek (input_plugin_t *this_gen,
   switch (origin) {
   case SEEK_SET:
     dist = offset / VCDSECTORSIZE;
-    this->cur_sector = start + dist;
+    this->cur_sec = start + dist;
     break;
   case SEEK_CUR:
 
     if (offset) 
-      LOG_MSG_STDERR(this->xine, _("input_vcd: SEEK_CUR not implemented for offset != 0\n"));
+      xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: SEEK_CUR not implemented for offset != 0\n");
 
-    sector_pos = this->cur_sector;
+    sector_pos = this->cur_sec;
 
     return sector_pos * VCDSECTORSIZE;
 
     break;
   default:
-    LOG_MSG_STDERR(this->xine, _("input_vcd: error seek to origin %d not implemented!\n"),
-	     origin);
+    xprintf (this->cls->xine, XINE_VERBOSITY_DEBUG, 
+	     "input_vcd: error seek to origin %d not implemented!\n", origin);
     return 0;
   }
 
@@ -799,32 +735,36 @@ static off_t vcd_plugin_seek (input_plugin_t *this_gen,
 }
 #endif
 
-/*
- *
- */
 #if defined (__linux__) || defined(__sun)
 static off_t vcd_plugin_get_length (input_plugin_t *this_gen) {
   vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
   struct cdrom_msf0       *end_msf, *start_msf;
   off_t len ;
-
-  start_msf = &this->tocent[this->cur_track].cdte_addr.msf;
-  end_msf   = &this->tocent[this->cur_track+1].cdte_addr.msf;
-
-  len = 75 - start_msf->frame;
-
-  if (start_msf->second<60)
-    len += (59 - start_msf->second) * 75;
-
-  if (end_msf->minute > start_msf->minute) {
-    len += (end_msf->minute - start_msf->minute-1) * 60 * 75;
-
-    len += end_msf->second * 60;
-
-    len += end_msf->frame ;
-  }
   
-  return len * VCDSECTORSIZE;
+  if(this->cls->total_tracks) {
+    
+    start_msf = (struct cdrom_msf0 *)
+      &this->cls->tocent[this->cur_track].cdte_addr.msf;
+    end_msf   = (struct cdrom_msf0 *)
+      &this->cls->tocent[this->cur_track+1].cdte_addr.msf;
+    
+    len = 75 - start_msf->frame;
+    
+    if (start_msf->second<60)
+      len += (59 - start_msf->second) * 75;
+    
+    if (end_msf->minute > start_msf->minute) {
+      len += (end_msf->minute - start_msf->minute-1) * 60 * 75;
+      
+      len += end_msf->second * 60;
+      
+      len += end_msf->frame ;
+    }
+    
+    return len * VCDSECTORSIZE;
+  }
+
+  return (off_t) 0;
 }
 #elif defined (__FreeBSD__)
 static off_t vcd_plugin_get_length (input_plugin_t *this_gen) {
@@ -833,195 +773,227 @@ static off_t vcd_plugin_get_length (input_plugin_t *this_gen) {
 
 
   len = 
-    ntohl(this->tocent
+    ntohl(this->cls->tocent
 	  [this->cur_track+2 
-	  - this->tochdr.starting_track].addr.lba)
-    - ntohl(this->tocent
+	  - this->cls->tochdr.starting_track].addr.lba)
+    - ntohl(this->cls->tocent
 	    [this->cur_track+1 
-	    - this->tochdr.starting_track].addr.lba);
+	    - this->cls->tochdr.starting_track].addr.lba);
   
   return len * 2352; /*VCDSECTORSIZE;*/
 
 }
 #endif
 
-/*
- *
- */
 static off_t vcd_plugin_get_current_pos (input_plugin_t *this_gen){
 
 
   return (vcd_plugin_seek (this_gen, 0, SEEK_CUR));
 }
 
-/*
- *
- */
 static uint32_t vcd_plugin_get_capabilities (input_plugin_t *this_gen) {
   
-  return INPUT_CAP_SEEKABLE | INPUT_CAP_PREVIEW | INPUT_CAP_BLOCK | INPUT_CAP_AUTOPLAY | INPUT_CAP_GET_DIR;
+  return INPUT_CAP_SEEKABLE | INPUT_CAP_BLOCK ;
 }
 
-/*
- *
- */
 static uint32_t vcd_plugin_get_blocksize (input_plugin_t *this_gen) {
 
   return VCDSECTORSIZE;
 }
 
-/*
- *
- */
-#if defined (__linux__)
-static int vcd_plugin_eject_media (input_plugin_t *this_gen) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
-  int ret, status;
+static void vcd_plugin_dispose (input_plugin_t *this_gen ) {
 
-  if((this->fd = open(this->device, O_RDONLY|O_NONBLOCK)) > -1) {
-    if((status = ioctl(this->fd, CDROM_DRIVE_STATUS, CDSL_CURRENT)) > 0) {
-      switch(status) {
-      case CDS_TRAY_OPEN:
-	if((ret = ioctl(this->fd, CDROMCLOSETRAY)) != 0) {
-	  LOG_MSG(this->xine, _("input_vcd: CDROMCLOSETRAY failed: %s\n"), strerror(errno));  
-	}
-	break;
-      case CDS_DISC_OK:
-	if((ret = ioctl(this->fd, CDROMEJECT)) != 0) {
-	  LOG_MSG(this->xine, _("input_vcd: CDROMEJECT failed: %s\n"), strerror(errno));  
-	}
-	break;
-      }
-    }
-    else {
-      LOG_MSG(this->xine, _("input_vcd: CDROM_DRIVE_STATUS failed: %s\n"), 
-	      strerror(errno));
-      close(this->fd);
+  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
+
+  if (this->fd != -1)
+    close(this->fd);
+
+  free (this->mrl);
+  free (this);
+}
+
+static char* vcd_plugin_get_mrl (input_plugin_t *this_gen) {
+  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
+
+  return this->mrl;
+}
+
+static int vcd_plugin_get_optional_data (input_plugin_t *this_gen, 
+					 void *data, int data_type) {
+
+  return INPUT_OPTIONAL_UNSUPPORTED;
+}
+
+static int vcd_plugin_open (input_plugin_t *this_gen) {
+  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
+  vcd_input_class_t  *cls = this->cls;
+  char               *filename;
+  int                 fd;
+
+  fd = open (cls->device, O_RDONLY|O_EXCL);
+  if (fd == -1) {
+    return 0;
+  }
+
+  this->fd     = fd;
+
+  if (input_vcd_read_toc (this->cls, this->fd)) {
+    return 0;
+  }
+
+  filename = (char *) &this->mrl[5];
+  while (*filename == '/') filename++;
+
+  if (sscanf (filename, "%d", &this->cur_track) != 1) {
+    xprintf (cls->xine, XINE_VERBOSITY_LOG, 
+	     _("input_vcd: malformed MRL. Use vcdo:/<track #>\n"));
+    return 0;
+  }
+
+  if (this->cur_track>=this->cls->total_tracks) {
+    xprintf (cls->xine, XINE_VERBOSITY_LOG, 
+	     _("input_vcd: invalid track %d (valid range: 0 .. %d)\n"),
+	     this->cur_track, this->cls->total_tracks-1);
+    return 0;
+  }
+
+#if defined (__linux__) || defined(__sun)
+  this->cur_min   = this->cls->tocent[this->cur_track].cdte_addr.msf.minute;
+  this->cur_sec   = this->cls->tocent[this->cur_track].cdte_addr.msf.second;
+  this->cur_frame = this->cls->tocent[this->cur_track].cdte_addr.msf.frame;
+#elif defined (__FreeBSD__)
+  {
+    int bsize = 2352;
+    if (ioctl (this->fd, CDRIOCSETBLOCKSIZE, &bsize) == -1) {
+      xprintf (cls->xine, XINE_VERBOSITY_DEBUG, "input_vcd: error in CDRIOCSETBLOCKSIZE %d\n", errno);
       return 0;
     }
-  }
-
-  close(this->fd);
   
-  return 1;
-}
-#elif defined (__FreeBSD__)
-static int vcd_plugin_eject_media (input_plugin_t *this_gen) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
-  int fd;
-
-  if ((fd = open(this->device, O_RDONLY|O_NONBLOCK)) > -1) {
-    if (ioctl(fd, CDIOCALLOW) == -1) {
-      perror("ioctl(cdromallow)");
-    } else {
-      if (ioctl(fd, CDIOCEJECT) == -1) {
-	perror("ioctl(cdromeject)");
-      }
-    }
-    close(fd);
+    this->cur_sec = 
+      ntohl(this->cls->tocent
+	    [this->cur_track+1 - this->cls->tochdr.starting_track].addr.lba);
+    
   }
-  
-  return 1;
-}
-#elif defined (__sun)
-static int vcd_plugin_eject_media (input_plugin_t *this_gen) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
-  int fd, ret;
-
-  if ((fd = open(this->device, O_RDONLY|O_NONBLOCK)) > -1) {
-    if ((ret = ioctl(fd, CDROMEJECT)) != 0) {
-      LOG_MSG(this->xine, _("input_vcd: CDROMEJECT failed: %s\n"), strerror(errno));  
-    }
-    close(fd);
-  }
-
-  return 1;
-}
 #endif
+  
+  return 1;
+}
 
-/*
- *
- */
-static void vcd_plugin_close (input_plugin_t *this_gen) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
+static input_plugin_t *vcd_class_get_instance (input_class_t *cls_gen, xine_stream_t *stream, 
+				    const char *data) {
 
-  close(this->fd);
-  this->fd = -1;
+  vcd_input_class_t  *cls = (vcd_input_class_t *) cls_gen; 
+  vcd_input_plugin_t *this;
+  char               *mrl = strdup(data);
+
+  if (strncasecmp (mrl, "vcdo:/",6)) {
+    free (mrl);
+    return 0;
+  }
+    
+  this = (vcd_input_plugin_t *) xine_xmalloc(sizeof(vcd_input_plugin_t));
+
+  this->stream = stream;
+  this->mrl    = mrl;
+  this->fd     = -1;
+
+  this->input_plugin.open              = vcd_plugin_open;
+  this->input_plugin.get_capabilities  = vcd_plugin_get_capabilities;
+  this->input_plugin.read              = vcd_plugin_read;
+  this->input_plugin.read_block        = vcd_plugin_read_block;
+  this->input_plugin.seek              = vcd_plugin_seek;
+  this->input_plugin.get_current_pos   = vcd_plugin_get_current_pos;
+  this->input_plugin.get_length        = vcd_plugin_get_length;
+  this->input_plugin.get_blocksize     = vcd_plugin_get_blocksize;
+  this->input_plugin.get_mrl           = vcd_plugin_get_mrl;
+  this->input_plugin.get_optional_data = vcd_plugin_get_optional_data;
+  this->input_plugin.dispose           = vcd_plugin_dispose;
+  this->input_plugin.input_class       = cls_gen;
+  this->cls                            = cls;
+
+  return &this->input_plugin;
 }
 
 /*
- *
+ * vcd input plugin class stuff
  */
-static void vcd_plugin_stop (input_plugin_t *this_gen) {
-  vcd_plugin_close(this_gen);
+
+static char *vcd_class_get_description (input_class_t *this_gen) {
+  return _("Video CD input plugin");
 }
 
-/*
- *
- */
-static char *vcd_plugin_get_description (input_plugin_t *this_gen) {
-  return _("vcd device input plugin as shipped with xine");
+static const char *vcd_class_get_identifier (input_class_t *this_gen) {
+  return "vcdo";
 }
 
-/*
- *
- */
-static char *vcd_plugin_get_identifier (input_plugin_t *this_gen) {
-  return "VCD";
+static void vcd_class_dispose (input_class_t *this_gen) {
+
+  vcd_input_class_t  *this = (vcd_input_class_t *) this_gen;
+  int                 i;
+
+  for (i = 0; i < 100; i++)
+    free (this->filelist[i]);
+
+  free (this->mrls);
+  free (this);
 }
 
-/*
- *
- */
-static mrl_t **vcd_plugin_get_dir (input_plugin_t *this_gen, 
-				   char *filename, int *nEntries) {
+static int vcd_class_eject_media (input_class_t *this_gen) {
 
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
-  int i;
+  vcd_input_class_t  *this = (vcd_input_class_t *) this_gen;
 
-  *nEntries = 0;
+  return media_eject_media (this->xine, this->device);
+}
+
+static xine_mrl_t **vcd_class_get_dir (input_class_t *this_gen, const char *filename, 
+				       int *num_files) {
+  vcd_input_class_t *this = (vcd_input_class_t *) this_gen;
+
+  int i, fd;
+
+  *num_files = 0;
 
   if (filename)
     return NULL;
 
   
-  this->fd = open (this->device, O_RDONLY);
+  fd = open (this->device, O_RDONLY|O_EXCL);
 
-  if (this->fd == -1) {
-    LOG_MSG_STDERR(this->xine, _("unable to open %s: %s.\n"), this->device, strerror(errno));
+  if (fd == -1) {
+    xprintf (this->xine, XINE_VERBOSITY_LOG,
+	     _("unable to open %s: %s.\n"), this->device, strerror(errno));
     return NULL;
   }
 
-  if (input_vcd_read_toc (this)) {
-    close (this->fd);
-    this->fd = -1;
+  if (input_vcd_read_toc (this, fd)) {
+    close (fd);
+    fd = -1;
 
-    LOG_MSG(this->xine, _("vcd_read_toc failed\n"));
+    xprintf (this->xine, XINE_VERBOSITY_DEBUG, "vcd_read_toc failed\n");
 
     return NULL;
   }
 
-  close (this->fd);
-  this->fd = -1;
+  close (fd);
+  fd = -1;
 
-  *nEntries = this->total_tracks - 1;
-  
+  *num_files = this->total_tracks - 1;
   /* printf ("%d tracks\n", this->total_tracks); */
 
   for (i=1; i<this->total_tracks; i++) { /* FIXME: check if track 0 contains valid data */
     char mrl[1024];
     
     memset(&mrl, 0, sizeof (mrl));
-    sprintf(mrl, "vcd://%d",i);
+    sprintf(mrl, "vcdo:/%d",i);
     
     if((i-1) >= this->mrls_allocated_entries) {
       ++this->mrls_allocated_entries;
       /* note: 1 extra pointer for terminating NULL */
-      this->mrls = realloc(this->mrls, (this->mrls_allocated_entries+1) * sizeof(mrl_t*));
-      this->mrls[(i-1)] = (mrl_t *) xine_xmalloc(sizeof(mrl_t));
+      this->mrls = realloc(this->mrls, (this->mrls_allocated_entries+1) * sizeof(xine_mrl_t*));
+      this->mrls[(i-1)] = (xine_mrl_t *) xine_xmalloc(sizeof(xine_mrl_t));
     }
     else {
-      memset(this->mrls[(i-1)], 0, sizeof(mrl_t));
+      memset(this->mrls[(i-1)], 0, sizeof(xine_mrl_t));
     }
     
     if(this->mrls[(i-1)]->mrl) {
@@ -1038,7 +1010,6 @@ static mrl_t **vcd_plugin_get_dir (input_plugin_t *this_gen,
     this->mrls[i-1]->type = (0 | mrl_vcd);
 
     /* hack */
-    this->cur_track = i;
     this->mrls[i-1]->size = vcd_plugin_get_length ((input_plugin_t *) this);
   }
 
@@ -1046,45 +1017,43 @@ static mrl_t **vcd_plugin_get_dir (input_plugin_t *this_gen,
   /*
    * Freeing exceeded mrls if exists.
    */
-  while(this->mrls_allocated_entries > *nEntries) {
+  while(this->mrls_allocated_entries > *num_files) {
     MRL_ZERO(this->mrls[this->mrls_allocated_entries - 1]);
     free(this->mrls[this->mrls_allocated_entries--]);
   }
 
-  this->mrls[*nEntries] = NULL;
+  this->mrls[*num_files] = NULL;
   
   return this->mrls;
 }
 
-/*
- *
- */
-static char **vcd_plugin_get_autoplay_list (input_plugin_t *this_gen, 
-					    int *nFiles) {
+static char ** vcd_class_get_autoplay_list (input_class_t *this_gen, int *num_files) {
 
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
-  int i;
+  vcd_input_class_t  *this = (vcd_input_class_t *) this_gen; 
+  int i, fd;
 
-  this->fd = open (this->device, O_RDONLY);
 
-  if (this->fd == -1) {
-    LOG_MSG_STDERR(this->xine, _("unable to open %s: %s.\n"), this->device, strerror(errno));
+  fd = open (this->device, O_RDONLY|O_EXCL);
+
+  if (fd == -1) {
+    xprintf (this->xine, XINE_VERBOSITY_LOG,
+	     _("input_vcd: unable to open %s: %s.\n"), this->device, strerror(errno));
     return NULL;
   }
 
-  if (input_vcd_read_toc (this)) {
-    close (this->fd);
-    this->fd = -1;
+  if (input_vcd_read_toc (this, fd)) {
+    close (fd);
+    fd = -1;
 
-    LOG_MSG(this->xine, _("vcd_read_toc failed\n"));
+    xprintf (this->xine, XINE_VERBOSITY_DEBUG, "input_vcd: vcd_read_toc failed\n");
 
     return NULL;
   }
 
-  close (this->fd);
-  this->fd = -1;
+  close (fd);
+  fd = -1;
 
-  *nFiles = this->total_tracks - 1;
+  *num_files = this->total_tracks - 1;
   
   /* printf ("%d tracks\n", this->total_tracks); */
 
@@ -1093,7 +1062,7 @@ static char **vcd_plugin_get_autoplay_list (input_plugin_t *this_gen,
     if(this->filelist[i - 1] == NULL)
       this->filelist[i - 1] = (char *) realloc(this->filelist[i - 1], sizeof(char *) * 256);
 
-    sprintf (this->filelist[i - 1], "vcd://%d",i);
+    sprintf (this->filelist[i - 1], "vcdo:/%d",i);
     /* printf ("list[%d] : %d %s\n", i, this->filelist[i-1], this->filelist[i-1]);   */
   }
 
@@ -1103,93 +1072,46 @@ static char **vcd_plugin_get_autoplay_list (input_plugin_t *this_gen,
   return this->filelist;
 }
 
-/*
- *
- */
-static char* vcd_plugin_get_mrl (input_plugin_t *this_gen) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
+static void *init_class (xine_t *xine, void *data) {
 
-  return this->mrl;
-}
-
-/*
- *
- */
-static int vcd_plugin_get_optional_data (input_plugin_t *this_gen, 
-					 void *data, int data_type) {
-
-  return INPUT_OPTIONAL_UNSUPPORTED;
-}
-
-static void vcd_plugin_dispose (input_plugin_t *this_gen ) {
-  vcd_input_plugin_t *this = (vcd_input_plugin_t *) this_gen;
-  int i;
-
-  for (i = 0; i < 100; i++)
-    free (this->filelist[i]);
-
-  free (this->mrls);
-  free (this);
-}
-
-/*
- *
- */
-input_plugin_t *init_input_plugin (int iface, xine_t *xine) {
-
-  vcd_input_plugin_t *this;
-  config_values_t    *config;
+  vcd_input_class_t  *this;
+  config_values_t    *config = xine->config;
   int                 i;
 
-  if (iface != 8) {
-    LOG_MSG(xine, 
-	    _("vcd input plugin doesn't support plugin API version %d.\n"
-	      "PLUGIN DISABLED.\n"
-	      "This means there's a version mismatch between xine and this input"
-	      "plugin.\nInstalling current input plugins should help.\n"),
-	    iface);
-    return NULL;
-  }
-    
-  this       = (vcd_input_plugin_t *) xine_xmalloc(sizeof(vcd_input_plugin_t));
-  config     = xine->config;
-  this->xine = xine;
+  this = (vcd_input_class_t *) xine_xmalloc (sizeof (vcd_input_class_t));
+
+  this->xine   = xine;
+
+  this->input_class.get_instance       = vcd_class_get_instance;
+  this->input_class.get_identifier     = vcd_class_get_identifier;
+  this->input_class.get_description    = vcd_class_get_description;
+  this->input_class.get_dir            = vcd_class_get_dir;
+  this->input_class.get_autoplay_list  = vcd_class_get_autoplay_list;
+  this->input_class.dispose            = vcd_class_dispose;
+  this->input_class.eject_media        = vcd_class_eject_media;
+
+  this->device = config->register_string (config, "media.vcd.device", CDROM,
+					  _("device used for VCD playback"),
+					  _("The path to the device, usually a CD or DVD drive, "
+					    "you intend to play your VideoCDs with."),
+					  10, device_change_cb, (void *)this);
+
+  this->mrls = (xine_mrl_t **) xine_xmalloc(sizeof(xine_mrl_t*));
+  this->mrls_allocated_entries = 0;
 
   for (i = 0; i < 100; i++) {
     this->filelist[i]       = (char *) xine_xmalloc(sizeof(char *) * 256);
   }
   
-  this->input_plugin.interface_version = INPUT_PLUGIN_IFACE_VERSION;
-  this->input_plugin.get_capabilities  = vcd_plugin_get_capabilities;
-  this->input_plugin.open              = vcd_plugin_open;
-  this->input_plugin.read              = vcd_plugin_read;
-  this->input_plugin.read_block        = vcd_plugin_read_block;
-  this->input_plugin.seek              = vcd_plugin_seek;
-  this->input_plugin.get_current_pos   = vcd_plugin_get_current_pos;
-  this->input_plugin.get_length        = vcd_plugin_get_length;
-  this->input_plugin.get_blocksize     = vcd_plugin_get_blocksize;
-  this->input_plugin.eject_media       = vcd_plugin_eject_media;
-  this->input_plugin.close             = vcd_plugin_close;
-  this->input_plugin.stop              = vcd_plugin_stop;
-  this->input_plugin.get_identifier    = vcd_plugin_get_identifier;
-  this->input_plugin.get_description   = vcd_plugin_get_description;
-  this->input_plugin.get_dir           = vcd_plugin_get_dir;
-  this->input_plugin.get_mrl           = vcd_plugin_get_mrl;
-  this->input_plugin.get_autoplay_list = vcd_plugin_get_autoplay_list;
-  this->input_plugin.get_optional_data = vcd_plugin_get_optional_data;
-  this->input_plugin.dispose           = vcd_plugin_dispose;
-  this->input_plugin.is_branch_possible= NULL;
-  
-  this->device = config->register_string(config, "input.vcd_device", CDROM,
-					 _("path to your local vcd device file"),
-					 NULL, device_change_cb, (void *)this);
-
-  this->mrls = (mrl_t **) xine_xmalloc(sizeof(mrl_t*));
-  this->mrls_allocated_entries = 0;
-
-  this->fd      = -1;
-  this->mrl     = NULL;
-  this->config  = config;
-  
-  return (input_plugin_t *) this;
+  return this;
 }
+
+/*
+ * exported plugin catalog entry
+ */
+
+plugin_info_t xine_plugin_info[] = {
+  /* type, API, "name", version, special_info, init_function */  
+  { PLUGIN_INPUT, 15, "VCDO", XINE_VERSION_CODE, NULL, init_class },
+  { PLUGIN_NONE, 0, "", 0, NULL, NULL }
+};
